@@ -48,7 +48,7 @@ public final class ConcurrentLogbackMDCAdapter extends LogbackMDCAdapter {
   // Initially the contents of the thread local in parent and child threads
   // reference the same map. However, as soon as a thread invokes the put()
   // method, the maps diverge as they should.
-  final InheritableThreadLocal<Map<String, String>> copyOnInheritThreadLocal = new InheritableThreadLocal<Map<String, String>>();
+  final ThreadLocal<Map<String, String>> copyOnThreadLocal = new ThreadLocal<Map<String, String>>();
 
   /**
    * Put a context value (the <code>val</code> parameter) as identified with the
@@ -67,11 +67,11 @@ public void put(String key, String val) throws IllegalArgumentException {
       throw new IllegalArgumentException("key cannot be null");
     }
 
-    Map<String, String> oldMap = copyOnInheritThreadLocal.get();
+    Map<String, String> oldMap = copyOnThreadLocal.get();
 
     if (oldMap == null) {
       Map<String, String> newMap = new ConcurrentHashMap<String, String>();
-      copyOnInheritThreadLocal.set(newMap);
+      copyOnThreadLocal.set(newMap);
       newMap.put(key, val);
     } else {
       oldMap.put(key, val);
@@ -87,7 +87,7 @@ public void remove(String key) {
     if (key == null) {
       return;
     }
-    Map<String, String> oldMap = copyOnInheritThreadLocal.get();
+    Map<String, String> oldMap = copyOnThreadLocal.get();
     if (oldMap == null) {
         return;
     }
@@ -101,7 +101,7 @@ public void remove(String key) {
    */
   @Override
 public void clear() {
-    copyOnInheritThreadLocal.remove();
+      copyOnThreadLocal.remove();
   }
 
   /**
@@ -120,7 +120,7 @@ public String get(String key) {
    */
   @Override
 public Map<String, String> getPropertyMap() {
-    return copyOnInheritThreadLocal.get();
+    return copyOnThreadLocal.get();
   }
 
   /**
@@ -138,7 +138,7 @@ public Map<String, String> getPropertyMap() {
    */
   @Override
 public Map getCopyOfContextMap() {
-    Map<String, String> hashMap = copyOnInheritThreadLocal.get();
+    Map<String, String> hashMap = copyOnThreadLocal.get();
     return hashMap == null ? null : new HashMap<String, String>(hashMap);
   }
 
@@ -148,6 +148,6 @@ public Map getCopyOfContextMap() {
     Map<String, String> newMap = new ConcurrentHashMap<String, String>(contextMap);
 
     // the newMap replaces the old one for serialisation's sake
-    copyOnInheritThreadLocal.set(newMap);
+    copyOnThreadLocal.set(newMap);
   }
 }
